from emodpy.campaign.common import ValueMap, CommonInterventionParameters
from emodpy.campaign.base_intervention import IndividualIntervention, _validate_value_range, _set_event
from emodpy.campaign.waning_config import AbstractWaningConfig
from emodpy.utils.emod_enum import NodeSelectionType, VaccineType, EventOrConfig
from emodpy.utils.distributions import BaseDistribution

from emod_api import campaign as api_campaign

from typing import Union


class BroadcastEvent(IndividualIntervention):
    """
    Create a new BroadcastEvent intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        broadcast_event(str, required):
            The name of the event to be broadcasted. See :doc:`emod-malaria:emod/parameter-campaign-event-list`
            for possible values, or create custom values using **Custom_Individual_Events**.

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: disqualifying_properties, new_property_value, intervention_name, dont_allow_duplicates.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 broadcast_event: str,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'BroadcastEvent', common_intervention_parameters)

        self.intervention.Broadcast_Event = _set_event(broadcast_event, 'broadcast_event', campaign, False)

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the BroadcastEvent intervention.')


class BroadcastEventToOtherNodes(IndividualIntervention):
    """
    Create a new BroadcastEventToOtherNodes intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

         broadcast_event(str, required):
            The name of the event to broadcast to the people of 'nearby' nodes. For example, if a house is
            found to have malarai, broadcast an event to the people in the nearby houses so that they can get
            treatment. See :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values.

         node_selection_type('NodeSelectionType', optional):
            The method by which to select nodes to receive the event. Possible values are:
            * DISTANCE_ONLY - Nodes located within the distance specified by **Max_Distance_To_Other_Nodes_Km**
            are selected.
            * MIGRATION_NODES_ONLY - Nodes that are local, regional, or connected in the migration file are
            selected.
            * DISTANCE_AND_MIGRATION - Nodes are selected using DISTANCE_ONLY and MIGRATION_NODES_ONLY criteria.
            Default value: DISTANCE_ONLY

        max_distance_to_other_nodes_km(float, optional):
            The maximum distance, in kilometers, to the destination node for the node to be selected. The
            location values used are those entered in the demographics file. Used only if
            **Node_Selection_Type** is either DISTANCE_ONLY or DISTANCE_AND_MIGRATION.
            Minimum value: 0
            Maximum value: 3.40282e+38
            Default value: 3.40282e+38

        include_my_node(bool, optional):
            Set to true (1) to broadcast the event to the current node.
            Default value: True

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 broadcast_event: str,
                 node_selection_type: 'NodeSelectionType' = NodeSelectionType.DISTANCE_ONLY,
                 max_distance_to_other_nodes_km: float = 3.40282e+38,
                 include_my_node: bool = False,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'BroadcastEventToOtherNodes', common_intervention_parameters)

        self.intervention.Node_Selection_Type = node_selection_type
        if node_selection_type in [NodeSelectionType.DISTANCE_ONLY, NodeSelectionType.DISTANCE_AND_MIGRATION]:
            self.intervention.Max_Distance_To_Other_Nodes_Km = _validate_value_range(max_distance_to_other_nodes_km, 'max_distance_to_other_nodes_km', 0, 3.40282e+38, float)
        else:
            if max_distance_to_other_nodes_km != 3.40282e+38:
                raise ValueError(f"max_distance_to_other_nodes_km is used when node_selection_type is set to "
                                 f"DISTANCE_ONLY or DISTANCE_AND_MIGRATION, not {node_selection_type}.")
            self.intervention.pop('Max_Distance_To_Other_Nodes_Km')
        self.intervention.Include_My_Node = include_my_node
        self.intervention.Event_Trigger = _set_event(broadcast_event, 'broadcast_event', campaign, False)

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the BroadcastEventToOtherNodes intervention.')


class ControlledVaccine(IndividualIntervention):
    """
    Create a new ControlledVaccine intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        waning_config(AbstractWaningConfig, required):
            The configuration of the vaccine's efficacy and waning over time. Specify how this effect decays
            over time using one of the Waning Config classes in emodpy.campaign.waninng_config.

        vaccine_type(VaccineType, optional):
            The type of vaccine to distribute in a vaccine intervention. Possible values are:
            * Generic - The vaccine can reduce transmission, acquisition, and mortality.
            * TransmissionBlocking - The vaccine will reduce pathogen transmission.
            * AcquisitionBlocking - The vaccine will reduce the acquisition of the pathogen by reducing the
            force of infection experienced by the vaccinated individual.
            * MortalityBlocking - The vaccine reduces the disease-mortality rate of a vaccinated individual.
            Default value: Generic

        vaccine_take(float, optional):
            The rate at which delivered vaccines will successfully stimulate an immune response and achieve the
            desired efficacy. For example, if it is set to 0.9, there will be a 90 percent chance that the
            vaccine will start with the specified efficacy, and a 10 percent chance that it will have no
            efficacy at all.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        expired_event_trigger(str, optional):
            The name of the event to be broadcast when the intervention is distributed to an individual. See
            :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values. Custom events can be
            distributed if they are defined in the config parameter **Custom_Individual_Events**.
            Default value: None

        efficacy_is_multiplicative(bool, optional):
            The overall vaccine efficacy when individuals receive more than one vaccine. When set to true (1),
            the vaccine efficacies are multiplied together; when set to false (0), the efficacies are additive.
            Default value: True

        duration_to_wait_before_revaccination(float, optional):
            The length of time, in days, to wait before revaccinating an individual. After this time has
            passed, the individual can be revaccinated. If the first vaccine has not expired, the individual
            can receive the effect from both doses of the vaccine.
            Minimum value: 0
            Maximum value: 3.40282e+38
            Default value: 3.40282e+38

        distributed_event_trigger(str, optional):
            The name of the event to be broadcast when the intervention is distributed to an individual. See
            :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values. Custom events can be
            distributed if they are defined in the config parameter **Custom_Individual_Events**.
            Default value: None

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 waning_config: AbstractWaningConfig,
                 vaccine_type: VaccineType = VaccineType.Generic,
                 vaccine_take: float = 1,
                 expired_event_trigger: str = None,
                 efficacy_is_multiplicative: bool = True,
                 duration_to_wait_before_revaccination: float = 3.40282e+38,
                 distributed_event_trigger: str = None,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'ControlledVaccine', common_intervention_parameters)
        if not isinstance(waning_config, AbstractWaningConfig):
            raise ValueError(f"waning_config must be an instance of AbstractWaningConfig, not {type(waning_config)}.")
        self.intervention.Waning_Config = waning_config.to_schema_dict(campaign)
        self.intervention.Vaccine_Type = vaccine_type
        self.intervention.Vaccine_Take = _validate_value_range(vaccine_take, 'vaccine_take', 0, 1, float)
        self.intervention.Expired_Event_Trigger = _set_event(expired_event_trigger, 'expired_event_trigger', campaign, True)
        self.intervention.Efficacy_Is_Multiplicative = efficacy_is_multiplicative
        self.intervention.Duration_To_Wait_Before_Revaccination = _validate_value_range(duration_to_wait_before_revaccination, 'duration_to_wait_before_revaccination', 0, 3.40282e+38, float)
        self.intervention.Distributed_Event_Trigger = _set_event(distributed_event_trigger, 'distributed_event_trigger', campaign, True)


class DelayedIntervention(IndividualIntervention):
    """
    Create a new DelayedIntervention intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        actual_individualintervention_configs(IndividualIntervention, required):
            An array of nested interventions to be distributed at the end of the delay period, to the covered
            fraction of the population.

        delay_period_distribution(BaseDistribution, required):
            The distribution type to use for assigning the delay period for distributing interventions. Each
            assigned value is a random draw from the distribution. Please use the following distribution classes
            from emodpy.utils.distributions to define the distribution:
            * ConstantDistribution
            * UniformDistribution
            * GaussianDistribution
            * ExponentialDistribution
            * PoissonDistribution
            * LogNormalDistribution
            * DualConstantDistribution
            * WeibullDistribution
            * DualExponentialDistribution

        coverage(float, optional):
            The proportion of individuals who receive this intervention that actually receive the configured
            interventions/events. This can be useful to target the intervention to a subset of the people that
            got the delay. To target the subgroup more specifically, distribute a **BroadcastEvent** and use a
            **NodeLevelHealthTriggeredIV** to distributes the desired intervention.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 actual_individualintervention_configs: IndividualIntervention,
                 delay_period_distribution: BaseDistribution,
                 coverage: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'DelayedIntervention', common_intervention_parameters)

        self.intervention.Actual_IndividualIntervention_Configs = actual_individualintervention_configs.intervention
        self.set_distribution(delay_period_distribution, 'Delay_Period')
        self.intervention.Coverage = _validate_value_range(coverage, 'coverage', 0, 1, float)

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the DelayedIntervention intervention.')


class IndividualImmunityChanger(IndividualIntervention):
    """
    Create a new IndividualImmunityChanger intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        prime_transmit(float, optional):
            Specifies the priming effect on transmission immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        prime_mortality(float, optional):
            Specifies the priming effect on mortality immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        prime_acquire(float, optional):
            Specifies the priming effect on acquisition immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_transmit(float, optional):
            Specifies the boosting effect on transmission immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_transmit(float, optional):
            Specifies how much transmission immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_mortality(float, optional):
            Specifies how much mortality immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_acquire(float, optional):
            Specifies how much acquisition immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_mortality(float, optional):
            Specifies the boosting effect on mortality immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_acquire(float, optional):
            Specifies the boosting effect on acquisition immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 1 common
            parameters: cost.
            The following parameters are not valid for this intervention:
            intervention_name
            dont_allow_duplicates
            new_property_value
            disqualifying_properties
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 prime_transmit: float = 0,
                 prime_mortality: float = 0,
                 prime_acquire: float = 0,
                 boost_transmit: float = 0,
                 boost_threshold_transmit: float = 0,
                 boost_threshold_mortality: float = 0,
                 boost_threshold_acquire: float = 0,
                 boost_mortality: float = 0,
                 boost_acquire: float = 0,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'IndividualImmunityChanger', common_intervention_parameters)

        self.intervention.Prime_Transmit = _validate_value_range(prime_transmit, 'prime_transmit', 0, 1, float)
        self.intervention.Prime_Mortality = _validate_value_range(prime_mortality, 'prime_mortality', 0, 1, float)
        self.intervention.Prime_Acquire = _validate_value_range(prime_acquire, 'prime_acquire', 0, 1, float)
        self.intervention.Boost_Transmit = _validate_value_range(boost_transmit, 'boost_transmit', 0, 1, float)
        self.intervention.Boost_Threshold_Transmit = _validate_value_range(boost_threshold_transmit, 'boost_threshold_transmit', 0, 1, float)
        self.intervention.Boost_Threshold_Mortality = _validate_value_range(boost_threshold_mortality, 'boost_threshold_mortality', 0, 1, float)
        self.intervention.Boost_Threshold_Acquire = _validate_value_range(boost_threshold_acquire, 'boost_threshold_acquire', 0, 1, float)
        self.intervention.Boost_Mortality = _validate_value_range(boost_mortality, 'boost_mortality', 0, 1, float)
        self.intervention.Boost_Acquire = _validate_value_range(boost_acquire, 'boost_acquire', 0, 1, float)

    def _set_intervention_name(self, intervention_name: str) -> None:
        raise ValueError('Intervention_Name is not a valid parameter for the IndividualImmunityChanger intervention.')

    def _set_dont_allow_duplicates(self, dont_allow_duplicates: bool) -> None:
        raise ValueError('Dont_Allow_Duplicates is not a valid parameter for the IndividualImmunityChanger intervention.')

    def _set_new_property_value(self, new_property_value: str) -> None:
        raise ValueError('New_Property_Value is not a valid parameter for the IndividualImmunityChanger intervention.')

    def _set_disqualifying_properties(self, disqualifying_properties: Union[dict[str, str], list[str]]) -> None:
        raise ValueError('Disqualifying_Properties is not a valid parameter for the IndividualImmunityChanger intervention.')


class IndividualNonDiseaseDeathRateModifier(IndividualIntervention):
    """
    Create a new IndividualNonDiseaseDeathRateModifier intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        duration_to_modifier(ValueMap, required):
            An array of Times and Values used to specify different modifiers over the duration of the
            intervention. Linear interpolation is used to find the values between time points. If the duration
            exceeds the max time, then the last modifier value will be used.

        expiration_event(str, optional):
            When the person stops using the intervention, this event will be broadcasted. See
            :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values or define custom events
            using config parameter **Custom_Individual_Events**.
            Default value: None

        expiration_duration_distribution(BaseDistribution, optional):
            For the distribution of each intervention, a randomly selected duration from this distribution will
            determine when the person stops using the intervention. This is independent of how long the
            intervention is effective. Please use the following distribution classes
            from emodpy.utils.distributions to define the distribution:
            * ConstantDistribution
            * UniformDistribution
            * GaussianDistribution
            * ExponentialDistribution
            * PoissonDistribution
            * LogNormalDistribution
            * DualConstantDistribution
            * WeibullDistribution
            * DualExponentialDistribution
            Default value: None

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 duration_to_modifier: ValueMap,
                 expiration_event: str = None,
                 expiration_duration_distribution: BaseDistribution = None,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'IndividualNonDiseaseDeathRateModifier', common_intervention_parameters)

        self.intervention.Duration_To_Modifier = duration_to_modifier.to_schema_dict(campaign)
        self.intervention.Expiration_Event = _set_event(expiration_event, 'expiration_event', campaign, True)
        if expiration_duration_distribution:
            self.set_distribution(expiration_duration_distribution, 'Expiration_Duration')


class MigrateIndividuals(IndividualIntervention):
    """
    Create a new MigrateIndividuals intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        nodeid_to_migrate_to(int, optional):
            The destination node ID for intervention-based migration.
            Minimum value: 0
            Maximum value: 4294970000.0
            Default value: 0

        is_moving(bool, optional):
            Set to true (1) to indicate the individual is permanently moving to a new home node for
            intervention-based migration. Once at the new home node, trips will be made with this node as the
            root (i.e. round trips come back to this node).
            Default value: True

        duration_before_leaving_distribution(BaseDistribution, optional):
            The distribution type to use for assigning the duration of time an individual waits before
            migrating to the destination node after intervention-based migration. Each assigned value is a
            random draw from the distribution. Please use the following distribution classes
            from emodpy.utils.distributions to define the distribution:
            * ConstantDistribution
            * UniformDistribution
            * GaussianDistribution
            * ExponentialDistribution
            * PoissonDistribution
            * LogNormalDistribution
            * DualConstantDistribution
            * WeibullDistribution
            * DualExponentialDistribution
            Default value: None

        duration_at_node_distribution(BaseDistribution, optional):
            The distribution type to use for assigning the duration of time an individual or family spends at a
            destination node after intervention-based migration. Each assigned value is a random draw from the
            distribution. Please use the following distribution classes
            from emodpy.utils.distributions to define the distribution:
            * ConstantDistribution
            * UniformDistribution
            * GaussianDistribution
            * ExponentialDistribution
            * PoissonDistribution
            * LogNormalDistribution
            * DualConstantDistribution
            * WeibullDistribution
            * DualExponentialDistribution
            Default value: None

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 nodeid_to_migrate_to: int = 0,
                 is_moving: bool = False,
                 duration_before_leaving_distribution: BaseDistribution = None,
                 duration_at_node_distribution: BaseDistribution = None,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'MigrateIndividuals', common_intervention_parameters)

        self.intervention.NodeID_To_Migrate_To = _validate_value_range(nodeid_to_migrate_to, 'nodeid_to_migrate_to', 0, 4294970000.0, int)
        self.intervention.Is_Moving = is_moving
        if duration_before_leaving_distribution:
            self.set_distribution(duration_before_leaving_distribution, 'Duration_Before_Leaving')
        if duration_at_node_distribution:
            self.set_distribution(duration_at_node_distribution, 'Duration_At_Node')

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the MigrateIndividuals intervention.')


class MultiEffectBoosterVaccine(IndividualIntervention):
    """
    Create a new MultiEffectBoosterVaccine intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        transmit_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine transmission. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        mortality_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine mortality. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        acquire_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine acquisition. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        vaccine_take(float, optional):
            The rate at which delivered vaccines will successfully stimulate an immune response and achieve the
            desired efficacy. For example, if it is set to 0.9, there will be a 90 percent chance that the
            vaccine will start with the specified efficacy, and a 10 percent chance that it will have no
            efficacy at all.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        prime_transmit(float, optional):
            Specifies the priming effect on transmission immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        prime_mortality(float, optional):
            Specifies the priming effect on mortality immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        prime_acquire(float, optional):
            Specifies the priming effect on acquisition immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_transmit(float, optional):
            Specifies the boosting effect on transmission immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_transmit(float, optional):
            Specifies how much transmission immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_mortality(float, optional):
            Specifies how much mortality immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_threshold_acquire(float, optional):
            Specifies how much acquisition immunity is required before the vaccine changes from a prime to a
            boost.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_mortality(float, optional):
            Specifies the boosting effect on mortality immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_acquire(float, optional):
            Specifies the boosting effect on acquisition immunity for naive individuals (without natural or
            vaccine-derived immunity) for a multi-effect booster vaccine. This does not replace current
            immunity, it builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 transmit_config: AbstractWaningConfig,
                 mortality_config: AbstractWaningConfig,
                 acquire_config: AbstractWaningConfig,
                 vaccine_take: float = 1,
                 prime_transmit: float = 0,
                 prime_mortality: float = 0,
                 prime_acquire: float = 0,
                 boost_transmit: float = 0,
                 boost_threshold_transmit: float = 0,
                 boost_threshold_mortality: float = 0,
                 boost_threshold_acquire: float = 0,
                 boost_mortality: float = 0,
                 boost_acquire: float = 0,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'MultiEffectBoosterVaccine', common_intervention_parameters)
        if not isinstance(transmit_config, AbstractWaningConfig):
            raise ValueError(f"transmit_config must be an instance of AbstractWaningConfig, not {type(transmit_config)}.")
        if not isinstance(mortality_config, AbstractWaningConfig):
            raise ValueError(f"mortality_config must be an instance of AbstractWaningConfig, not {type(mortality_config)}.")
        if not isinstance(acquire_config, AbstractWaningConfig):
            raise ValueError(f"acquire_config must be an instance of AbstractWaningConfig, not {type(acquire_config)}.")

        self.intervention.Transmit_Config = transmit_config.to_schema_dict(campaign)
        self.intervention.Mortality_Config = mortality_config.to_schema_dict(campaign)
        self.intervention.Acquire_Config = acquire_config.to_schema_dict(campaign)
        self.intervention.Vaccine_Take = _validate_value_range(vaccine_take, 'vaccine_take', 0, 1, float)
        self.intervention.Prime_Transmit = _validate_value_range(prime_transmit, 'prime_transmit', 0, 1, float)
        self.intervention.Prime_Mortality = _validate_value_range(prime_mortality, 'prime_mortality', 0, 1, float)
        self.intervention.Prime_Acquire = _validate_value_range(prime_acquire, 'prime_acquire', 0, 1, float)
        self.intervention.Boost_Transmit = _validate_value_range(boost_transmit, 'boost_transmit', 0, 1, float)
        self.intervention.Boost_Threshold_Transmit = _validate_value_range(boost_threshold_transmit, 'boost_threshold_transmit', 0, 1, float)
        self.intervention.Boost_Threshold_Mortality = _validate_value_range(boost_threshold_mortality, 'boost_threshold_mortality', 0, 1, float)
        self.intervention.Boost_Threshold_Acquire = _validate_value_range(boost_threshold_acquire, 'boost_threshold_acquire', 0, 1, float)
        self.intervention.Boost_Mortality = _validate_value_range(boost_mortality, 'boost_mortality', 0, 1, float)
        self.intervention.Boost_Acquire = _validate_value_range(boost_acquire, 'boost_acquire', 0, 1, float)


class MultiEffectVaccine(IndividualIntervention):
    """
    Create a new MultiEffectVaccine intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        transmit_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine transmission. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        mortality_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine mortality. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        acquire_config(AbstractWaningConfig, required):
            The configuration for multi-effect vaccine acquisition. Specify how this effect decays over time
            using one of the Waning Config classes in emodpy.campaign.waninng_config.

        vaccine_take(float, optional):
            The rate at which delivered vaccines will successfully stimulate an immune response and achieve the
            desired efficacy. For example, if it is set to 0.9, there will be a 90 percent chance that the
            vaccine will start with the specified efficacy, and a 10 percent chance that it will have no
            efficacy at all.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 transmit_config: AbstractWaningConfig,
                 mortality_config: AbstractWaningConfig,
                 acquire_config: AbstractWaningConfig,
                 vaccine_take: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'MultiEffectVaccine', common_intervention_parameters)
        if not isinstance(transmit_config, AbstractWaningConfig):
            raise ValueError(f"transmit_config must be an instance of AbstractWaningConfig, not {type(transmit_config)}.")
        if not isinstance(mortality_config, AbstractWaningConfig):
            raise ValueError(f"mortality_config must be an instance of AbstractWaningConfig, not {type(mortality_config)}.")
        if not isinstance(acquire_config, AbstractWaningConfig):
            raise ValueError(f"acquire_config must be an instance of AbstractWaningConfig, not {type(acquire_config)}.")

        self.intervention.Transmit_Config = transmit_config.to_schema_dict(campaign)
        self.intervention.Mortality_Config = mortality_config.to_schema_dict(campaign)
        self.intervention.Acquire_Config = acquire_config.to_schema_dict(campaign)
        self.intervention.Vaccine_Take = _validate_value_range(vaccine_take, 'vaccine_take', 0, 1, float)


class MultiInterventionDistributor(IndividualIntervention):
    """
    Create a new MultiInterventionDistributor intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        intervention_list(list[IndividualIntervention], required):
            The list of individual interventions that is distributed by **MultiInterventionDistributor**.

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 intervention_list: list[IndividualIntervention],
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'MultiInterventionDistributor', common_intervention_parameters)

        self.intervention.Intervention_List = [i.intervention for i in intervention_list]

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the MultiInterventionDistributor intervention.')


class OutbreakIndividual(IndividualIntervention):
    """
    Create a new OutbreakIndividual intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        incubation_period_override(int, optional):
            The incubation period, in days, that infected individuals will go through before becoming
            infectious. This value overrides the incubation period set in the configuration file. Note for HIV
            simulations only: If set to 0, infection is assumed to be part of an outbreak event and a random
            duration until death is determined. For all other simulation types: Set to -1 to honor the
            configuration parameter settings.
            Minimum value: -1
            Maximum value: 2147480000.0
            Default value: -1

        ignore_immunity(bool, optional):
            Individuals will be force-infected (with a specific strain) regardless of actual immunity level
            when set to true (1).
            Default value: True

        genome(int, optional):
            The genetic substrain ID of the outbreak infection. Together with **Antigen**, they are a unitary
            object representing a strain of infection, which allows for differentiation among infections.
            Minimum value: -1
            Maximum value: 16777200.0
            Default value: 0

        antigen(int, optional):
            The antigenic base strain ID of the outbreak infection.
            Minimum value: 0
            Maximum value: 10
            Default value: 0

    """

    def __init__(self,
                 campaign: api_campaign,
                 incubation_period_override: int = -1,
                 ignore_immunity: bool = True,
                 genome: int = 0,
                 antigen: int = 0):
        super().__init__(campaign, 'OutbreakIndividual')

        self.intervention.Incubation_Period_Override = _validate_value_range(incubation_period_override, 'incubation_period_override', -1, 2147480000.0, int)
        self.intervention.Ignore_Immunity = ignore_immunity
        self.intervention.Genome = _validate_value_range(genome, 'genome', -1, 16777200.0, int)
        self.intervention.Antigen = _validate_value_range(antigen, 'antigen', 0, 10, int)

    def _set_intervention_name(self, intervention_name: str) -> None:
        raise ValueError('Intervention_Name is not a valid parameter for the OutbreakIndividual intervention.')

    def _set_dont_allow_duplicates(self, dont_allow_duplicates: bool) -> None:
        raise ValueError('Dont_Allow_Duplicates is not a valid parameter for the OutbreakIndividual intervention.')

    def _set_new_property_value(self, new_property_value: str) -> None:
        raise ValueError('New_Property_Value is not a valid parameter for the OutbreakIndividual intervention.')

    def _set_disqualifying_properties(self, disqualifying_properties: Union[dict[str, str], list[str]]) -> None:
        raise ValueError('Disqualifying_Properties is not a valid parameter for the OutbreakIndividual intervention.')

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the OutbreakIndividual intervention.')


class PropertyValueChanger(IndividualIntervention):
    """
    Create a new PropertyValueChanger intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        target_property_value(str, required):
            The user-defined value of the individual property that will be assigned to the individual.

        target_property_key(str, required):
            The name of the individual property type whose value will be updated by the intervention.

        revert(float, optional):
            The number of days to keep the value (**Target_Property_Value**) of the property
            (**Target_Property_Key**) set by the intervenion for the individual. When the time has expired, the
            intervention will reset the property back to the value it had when the intervention was first
            applied.
            Minimum value: 0
            Maximum value: 3.40282e+38
            Default value: 0

        maximum_duration(float, optional):
            The maximum amount of time individuals have to move to a new group. This timing works in
            conjunction with **Daily_Probability**.
            Minimum value: -1
            Maximum value: 3.40282e+38
            Default value: 3.40282e+38

        daily_probability(float, optional):
            The daily probability that an individual's property value changes to the **Target_Property_Value**.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 target_property_value: str,
                 target_property_key: str,
                 revert: float = 0,
                 maximum_duration: float = 3.40282e+38,
                 daily_probability: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'PropertyValueChanger', common_intervention_parameters)

        self.intervention.Target_Property_Value = target_property_value
        self.intervention.Target_Property_Key = target_property_key
        self.intervention.Revert = _validate_value_range(revert, 'revert', 0, 3.40282e+38, float)
        self.intervention.Maximum_Duration = _validate_value_range(maximum_duration, 'maximum_duration', -1, 3.40282e+38, float)
        self.intervention.Daily_Probability = _validate_value_range(daily_probability, 'daily_probability', 0, 1, float)

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the PropertyValueChanger intervention.')


class SimpleBoosterVaccine(IndividualIntervention):
    """
    Create a new SimpleBoosterVaccine intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        waning_config(AbstractWaningConfig, required):
            The configuration of the vaccine's efficacy and waning over time. Specify how this effect decays
            over time using one of the Waning Config classes in emodpy.campaign.waninng_config.

        vaccine_type(VaccineType, optional):
            The type of vaccine to distribute in a vaccine intervention. Possible values are:
            * Generic - The vaccine can reduce transmission, acquisition, and mortality.
            * TransmissionBlocking - The vaccine will reduce pathogen transmission.
            * AcquisitionBlocking - The vaccine will reduce the acquisition of the pathogen by reducing the
            force of infection experienced by the vaccinated individual.
            * MortalityBlocking - The vaccine reduces the disease-mortality rate of a vaccinated individual.
            Default value: Generic

        vaccine_take(float, optional):
            The rate at which delivered vaccines will successfully stimulate an immune response and achieve the
            desired efficacy. For example, if it is set to 0.9, there will be a 90 percent chance that the
            vaccine will start with the specified efficacy, and a 10 percent chance that it will have no
            efficacy at all.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        prime_effect(float, optional):
            Specifies the priming effect on [acquisition/transmission/mortality] immunity for naive individuals
            (without natural or vaccine-derived immunity).
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        efficacy_is_multiplicative(bool, optional):
            The overall vaccine efficacy when individuals receive more than one vaccine. When set to true (1),
            the vaccine efficacies are multiplied together; when set to false (0), the efficacies are additive.
            Default value: True

        boost_threshold(float, optional):
            Specifies how much immunity is required before the vaccine changes from a priming effect to a
            boosting effect.
            Minimum value: 0
            Maximum value: 1
            Default value: 0

        boost_effect(float, optional):
            Specifies the boosting effect on [acquisition/transmission/mortality] immunity for previously
            exposed individuals (either natural or vaccine-derived). This does not replace current immunity, it
            builds multiplicatively on top of it.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 waning_config: AbstractWaningConfig,
                 vaccine_type: VaccineType = VaccineType.Generic,
                 vaccine_take: float = 1,
                 prime_effect: float = 1,
                 efficacy_is_multiplicative: bool = True,
                 boost_threshold: float = 0,
                 boost_effect: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'SimpleBoosterVaccine', common_intervention_parameters)
        if not isinstance(waning_config, AbstractWaningConfig):
            raise ValueError(f"waning_config must be an instance of AbstractWaningConfig, not {type(waning_config)}.")

        self.intervention.Waning_Config = waning_config.to_schema_dict(campaign)
        self.intervention.Vaccine_Type = vaccine_type
        self.intervention.Vaccine_Take = _validate_value_range(vaccine_take, 'vaccine_take', 0, 1, float)
        self.intervention.Prime_Effect = _validate_value_range(prime_effect, 'prime_effect', 0, 1, float)
        self.intervention.Efficacy_Is_Multiplicative = efficacy_is_multiplicative
        self.intervention.Boost_Threshold = _validate_value_range(boost_threshold, 'boost_threshold', 0, 1, float)
        self.intervention.Boost_Effect = _validate_value_range(boost_effect, 'boost_effect', 0, 1, float)


# todo: should we remove all event_or_config parameters and implicitly set it base on positive_diagnosis_config and positive_diagnosis_event?
class SimpleDiagnostic(IndividualIntervention):
    """
    Create a new SimpleDiagnostic intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        positive_diagnosis_config(IndividualIntervention, optional):
            The intervention distributed to individuals if they test positive. Only used when
            **Event_Or_Config** is set to Config.See :doc:`emod-malaria:emod/parameter-campaign-event-list`
            for possible values.
            Default Value: None

        treatment_fraction(float, optional):
            The fraction of positive diagnoses that are treated or given the **Positive_Diagnosis_Config** or
            **Positive_Diagnosis_Event** when used in campaign classes that trigger separate interventions or
            events upon a positive diagnosis.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        positive_diagnosis_event(str, optional):
            If the test is positive, this specifies an event that can trigger another intervention when the
            event occurs. Only used if **Event_Or_Config** is set to **Event**. See
            :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values.
            Default value: None

        event_or_config('EventOrConfig', optional):
            Specifies whether the current intervention distributes a nested intervention (the Config option) or
            an event will be broadcast which may trigger other interventions in the campaign file (the Event
            option). Possible values are: **Event**, **Config**
            Default value: Config

        enable_is_symptomatic(bool, optional):
            If true (1), requires an infection to be symptomatic to return a positive test.
            Default value: True

        days_to_diagnosis(float, optional):
            The number of days from diagnosis (which is done when the intervention is distributed) until a
            positive response is performed. The response to a negative diagnosis is done immediately when the
            diagnosis is made (at distribution of the intervention).
            Minimum value: 0
            Maximum value: 3.40282e+38
            Default value: 0

        base_specificity(float, optional):
            The specificity of the diagnostic. This sets the proportion of the time that individuals without
            the condition being tested receive a negative diagnostic test. When set to 1, the diagnostic always
            accurately reflects the lack of having the condition. When set to zero, then individuals who do not
            have the condition always receive a false-positive diagnostic test.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        base_sensitivity(float, optional):
            The sensitivity of the diagnostic. This sets the proportion of the time that individuals with the
            condition being tested receive a positive diagnostic test. When set to 1, the diagnostic always
            accurately reflects the condition. When set to zero, then individuals who have the condition always
            receive a false-negative diagnostic test.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 positive_diagnosis_config: IndividualIntervention = None,
                 treatment_fraction: float = 1,
                 positive_diagnosis_event: str = None,
                 event_or_config: 'EventOrConfig' = EventOrConfig.Config,
                 enable_is_symptomatic: bool = False,
                 days_to_diagnosis: float = 0,
                 base_specificity: float = 1,
                 base_sensitivity: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'SimpleDiagnostic', common_intervention_parameters)

        self.intervention.Event_Or_Config = event_or_config
        if event_or_config == EventOrConfig.Config:
            if positive_diagnosis_config is None:
                raise ValueError('positive_diagnosis_config is required when Event_Or_Config is set to Config.')
            if not isinstance(positive_diagnosis_config, IndividualIntervention):
                raise ValueError('positive_diagnosis_config must be a Intervention instance.')
            self.intervention.Positive_Diagnosis_Config = positive_diagnosis_config.intervention
            self.intervention.pop("Positive_Diagnosis_Event")
        elif event_or_config == EventOrConfig.Event:
            self.intervention.Positive_Diagnosis_Event = _set_event(positive_diagnosis_event,
                                                                    'positive_diagnosis_event', campaign, False)
            self.intervention.pop("Positive_Diagnosis_Config")
        else:
            raise ValueError('Event_Or_Config must be a member of the EventOrConfig enum.')

        self.intervention.Treatment_Fraction = _validate_value_range(treatment_fraction, 'treatment_fraction', 0, 1, float)
        self.intervention.Enable_Is_Symptomatic = enable_is_symptomatic
        self.intervention.Days_To_Diagnosis = _validate_value_range(days_to_diagnosis, 'days_to_diagnosis', 0, 3.40282e+38, float)
        self.intervention.Base_Specificity = _validate_value_range(base_specificity, 'base_specificity', 0, 1, float)
        self.intervention.Base_Sensitivity = _validate_value_range(base_sensitivity, 'base_sensitivity', 0, 1, float)


class SimpleHealthSeekingBehavior(IndividualIntervention):
    """
    Create a new SimpleHealthSeekingBehavior intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        actual_individualintervention_config(IndividualIntervention, optional):
            The configuration of an actual intervention sought. Selects a class for the intervention and
            configures the parameters specific for that intervention class. Only used if **Event_Or_Config** is
            set to 'Config'.
            Default value: None

        tendency(float, optional):
            The probability of seeking healthcare.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        single_use(bool, optional):
            If set to true (1), the health-seeking behavior gets used once and discarded. If set to false (0),
            it remains indefinitely.
            Default value: True

        event_or_config('EventOrConfig', optional):
            Specifies whether the current intervention distributes a nested intervention (the Config option) or
            an event will be broadcast which may trigger other interventions in the campaign file (the Event
            option). Possible values are: **Event**, **Config**
            Default value: Config

        actual_individualintervention_event(str, optional):
            The event of an actual intervention sought. Selects a class for the intervention and configures the
            parameters specific for that intervention class. See :doc:`emod-malaria:emod/parameter-campaign-event-list`
            for possible values. Used only if **Event_Or_Config** is set to 'Event'.
            Default value: None

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 4 common
            parameters: intervention_name, new_property_value, dont_allow_duplicates, disqualifying_properties.
            The following parameters are not valid for this intervention:
            cost
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 actual_individualintervention_config: IndividualIntervention = None,
                 tendency: float = 1,
                 single_use: bool = True,
                 event_or_config: 'EventOrConfig' = EventOrConfig.Config,
                 actual_individualintervention_event: str = None,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'SimpleHealthSeekingBehavior', common_intervention_parameters)

        self.intervention.Event_Or_Config = event_or_config
        if event_or_config == EventOrConfig.Config:
            if actual_individualintervention_config is None:
                raise ValueError('actual_individualintervention_config is required when Event_Or_Config is set to '
                                 'Config.')
            if not isinstance(actual_individualintervention_config, IndividualIntervention):
                raise ValueError('actual_individualintervention_config must be a Intervention instance.')
            self.intervention.Actual_IndividualIntervention_Config = actual_individualintervention_config.intervention
            self.intervention.pop("Actual_IndividualIntervention_Event")
        elif event_or_config == EventOrConfig.Event:
            self.intervention.Actual_IndividualIntervention_Event = _set_event(actual_individualintervention_event,
                                                                               'actual_individualintervention_event',
                                                                               campaign, False)
            self.intervention.pop("Actual_IndividualIntervention_Config")
        else:
            raise ValueError('Event_Or_Config must be a member of the EventOrConfig enum.')
        self.intervention.Tendency = _validate_value_range(tendency, 'tendency', 0, 1, float)
        self.intervention.Single_Use = single_use

    def _set_cost(self, cost: float) -> None:
        raise ValueError('Cost_To_Consumer is not a valid parameter for the SimpleHealthSeekingBehavior intervention.')


class SimpleVaccine(IndividualIntervention):
    """
    Create a new SimpleVaccine intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        waning_config(AbstractWaningConfig, required):
            The configuration of the vaccine's efficacy and waning over time. Specify how this effect decays
            over time using one of the Waning Config classes in emodpy.campaign.waning_config.

        vaccine_type(VaccineType, optional):
            The type of vaccine to distribute in a vaccine intervention. Possible values are:
            * Generic - The vaccine can reduce transmission, acquisition, and mortality.
            * TransmissionBlocking - The vaccine will reduce pathogen transmission.
            * AcquisitionBlocking - The vaccine will reduce the acquisition of the pathogen by reducing the
            force of infection experienced by the vaccinated individual.
            * MortalityBlocking - The vaccine reduces the disease-mortality rate of a vaccinated individual.
            Default value: Generic

        vaccine_take(float, optional):
            The rate at which delivered vaccines will successfully stimulate an immune response and achieve the
            desired efficacy. For example, if it is set to 0.9, there will be a 90 percent chance that the
            vaccine will start with the specified efficacy, and a 10 percent chance that it will have no
            efficacy at all.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        efficacy_is_multiplicative(bool, optional):
            The overall vaccine efficacy when individuals receive more than one vaccine. When set to true (1),
            the vaccine efficacies are multiplied together; when set to false (0), the efficacies are additive.
            Default value: True

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 waning_config: AbstractWaningConfig,
                 vaccine_type: VaccineType = VaccineType.Generic,
                 vaccine_take: float = 1,
                 efficacy_is_multiplicative: bool = True,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'SimpleVaccine', common_intervention_parameters)
        if not isinstance(waning_config, AbstractWaningConfig):
            raise ValueError(f"waning_config must be an instance of AbstractWaningConfig, not {type(waning_config)}.")

        self.intervention.Waning_Config = waning_config.to_schema_dict(campaign)
        self.intervention.Vaccine_Type = vaccine_type
        self.intervention.Vaccine_Take = _validate_value_range(vaccine_take, 'vaccine_take', 0, 1, float)
        self.intervention.Efficacy_Is_Multiplicative = efficacy_is_multiplicative


class StandardDiagnostic(IndividualIntervention):
    """
    Create a new StandardDiagnostic intervention.

    Args:
        campaign (api_campaign, required):
            An instance of the emod_api.campaign module.

        positive_diagnosis_config(IndividualIntervention, optional):
            The intervention distributed to individuals if they test positive. Only used when
            **Event_Or_Config** is set to Config.See :doc:`emod-malaria:emod/parameter-campaign-event-list`
            for possible values.
            Default value: None

        negative_diagnosis_config(IndividualIntervention, optional):
            If **Event_Or_Config** is set to 'Config', this parameter defines a new intervention to be given to
            the person on a negative test result.
            Default value: None

        treatment_fraction(float, optional):
            The fraction of positive diagnoses that are treated or given the **Positive_Diagnosis_Config** or
            **Positive_Diagnosis_Event** when used in campaign classes that trigger separate interventions or
            events upon a positive diagnosis.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        positive_diagnosis_event(str, optional):
            If the test is positive, this specifies an event that can trigger another intervention when the
            event occurs. Only used if **Event_Or_Config** is set to **Event**. See
            :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values.
            Default value: None

        negative_diagnosis_event(str, optional):
            If **Event_Or_Config** is set to **Event**, this parameter defines the event to be broadcasted on a
            negative test result. See :doc:`emod-malaria:emod/parameter-campaign-event-list` for possible values
            or define custom events by adding to the list for the configuration parameter **Custom_Individual_Event**.
            Default value: None

        event_or_config('EventOrConfig', optional):
            Specifies whether the current intervention distributes a nested intervention (the Config option) or
            an event will be broadcast which may trigger other interventions in the campaign file (the Event
            option). Possible values are: **Event**, **Config**
            Default value: Config

        enable_is_symptomatic(bool, optional):
            If true (1), requires an infection to be symptomatic to return a positive test.
            Default value: True

        days_to_diagnosis(float, optional):
            The number of days from diagnosis (which is done when the intervention is distributed) until a
            positive response is performed. The response to a negative diagnosis is done immediately when the
            diagnosis is made (at distribution of the intervention).
            Minimum value: 0
            Maximum value: 3.40282e+38
            Default value: 0

        base_specificity(float, optional):
            The specificity of the diagnostic. This sets the proportion of the time that individuals without
            the condition being tested receive a negative diagnostic test. When set to 1, the diagnostic always
            accurately reflects the lack of having the condition. When set to zero, then individuals who do not
            have the condition always receive a false-positive diagnostic test.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        base_sensitivity(float, optional):
            The sensitivity of the diagnostic. This sets the proportion of the time that individuals with the
            condition being tested receive a positive diagnostic test. When set to 1, the diagnostic always
            accurately reflects the condition. When set to zero, then individuals who have the condition always
            receive a false-negative diagnostic test.
            Minimum value: 0
            Maximum value: 1
            Default value: 1

        common_intervention_parameters (CommomInterventionParameters, optional):
            The CommonInterventionParameters object that Additional parameters that contains the 5 common
            parameters: cost, new_property_value, intervention_name, disqualifying_properties, dont_allow_duplicates.
            Default value: None
    """

    def __init__(self,
                 campaign: api_campaign,
                 positive_diagnosis_config: IndividualIntervention = None,
                 negative_diagnosis_config: IndividualIntervention = None,
                 treatment_fraction: float = 1,
                 positive_diagnosis_event: str = None,
                 negative_diagnosis_event: str = None,
                 event_or_config: 'EventOrConfig' = EventOrConfig.Config,
                 enable_is_symptomatic: bool = False,
                 days_to_diagnosis: float = 0,
                 base_specificity: float = 1,
                 base_sensitivity: float = 1,
                 common_intervention_parameters: CommonInterventionParameters = None):
        super().__init__(campaign, 'StandardDiagnostic', common_intervention_parameters)

        self.intervention.Event_Or_Config = event_or_config
        if event_or_config == EventOrConfig.Config:
            if positive_diagnosis_config is None:
                raise ValueError('positive_diagnosis_config is required when Event_Or_Config is set to Config.')
            if not isinstance(positive_diagnosis_config, IndividualIntervention):
                raise ValueError('positive_diagnosis_config must be a Intervention instance.')
            if negative_diagnosis_config is None:
                raise ValueError('negative_diagnosis_config is required when Event_Or_Config is set to Config.')
            if not isinstance(negative_diagnosis_config, IndividualIntervention):
                raise ValueError('negative_diagnosis_config must be a Intervention instance.')
            self.intervention.Positive_Diagnosis_Config = positive_diagnosis_config.intervention
            self.intervention.Negative_Diagnosis_Config = negative_diagnosis_config.intervention
            self.intervention.pop("Positive_Diagnosis_Event")
            self.intervention.pop("Negative_Diagnosis_Event")
        elif event_or_config == EventOrConfig.Event:
            self.intervention.Positive_Diagnosis_Event = _set_event(positive_diagnosis_event, 'positive_diagnosis_event', campaign, False)
            self.intervention.Negative_Diagnosis_Event = _set_event(negative_diagnosis_event, 'negative_diagnosis_event', campaign, False)
            self.intervention.pop("Positive_Diagnosis_Config")
            self.intervention.pop("Negative_Diagnosis_Config")
        else:
            raise ValueError('Event_Or_Config must be a member of the EventOrConfig enum.')
        self.intervention.Treatment_Fraction = _validate_value_range(treatment_fraction, 'treatment_fraction', 0, 1, float)
        self.intervention.Enable_Is_Symptomatic = enable_is_symptomatic
        self.intervention.Days_To_Diagnosis = _validate_value_range(days_to_diagnosis, 'days_to_diagnosis', 0, 3.40282e+38, float)
        self.intervention.Base_Specificity = _validate_value_range(base_specificity, 'base_specificity', 0, 1, float)
        self.intervention.Base_Sensitivity = _validate_value_range(base_sensitivity, 'base_sensitivity', 0, 1, float)
